@@ -71,6 +71,9 @@ static const struct datatype *datatypes[TYPE_MAX + 1] = {
 	[TYPE_BOOLEAN]		= &boolean_type,
 	[TYPE_IFNAME]		= &ifname_type,
 	[TYPE_IGMP_TYPE]	= &igmp_type_type,
+	[TYPE_TIME_DATE]	= &date_type,
+	[TYPE_TIME_HOUR]	= &hour_type,
+	[TYPE_TIME_DAY]		= &day_type,
 };
 
 const struct datatype *datatype_lookup(enum datatypes type)
@@ -213,8 +216,8 @@ void symbol_table_print(const struct symbol_table *tbl,
 			enum byteorder byteorder,
 			struct output_ctx *octx)
 {
+	unsigned int len = div_round_up(dtype->size, BITS_PER_BYTE);
 	const struct symbolic_constant *s;
-	unsigned int len = dtype->size / BITS_PER_BYTE;
 	uint64_t value;
 
 	for (s = tbl->symbols; s->identifier != NULL; s++) {
@@ -654,34 +657,13 @@ const struct datatype inet_protocol_type = {
 
 static void inet_service_print(const struct expr *expr, struct output_ctx *octx)
 {
-	struct sockaddr_in sin = { .sin_family = AF_INET };
-	char buf[NI_MAXSERV];
-	uint16_t port;
-	int err;
+	uint16_t port = mpz_get_be16(expr->value);
+	const struct servent *s = getservbyport(port, NULL);
 
-	sin.sin_port = mpz_get_be16(expr->value);
-	err = getnameinfo((struct sockaddr *)&sin, sizeof(sin), NULL, 0,
-			  buf, sizeof(buf), 0);
-	if (err != 0) {
-		nft_print(octx, "%u", ntohs(sin.sin_port));
-		return;
-	}
-	port = atoi(buf);
-	/* We got a TCP service name string, display it... */
-	if (htons(port) != sin.sin_port) {
-		nft_print(octx, "\"%s\"", buf);
-		return;
-	}
-
-	/* ...otherwise, this might be a UDP service name. */
-	err = getnameinfo((struct sockaddr *)&sin, sizeof(sin), NULL, 0,
-			  buf, sizeof(buf), NI_DGRAM);
-	if (err != 0) {
-		/* No service name, display numeric value. */
-		nft_print(octx, "%u", ntohs(sin.sin_port));
-		return;
-	}
-	nft_print(octx, "\"%s\"", buf);
+	if (s == NULL)
+		nft_print(octx, "%hu", ntohs(port));
+	else
+		nft_print(octx, "\"%s\"", s->s_name);
 }
 
 void inet_service_type_print(const struct expr *expr, struct output_ctx *octx)
@@ -1187,11 +1169,6 @@ const struct datatype *set_datatype_alloc(const struct datatype *orig_dtype,
 	return dtype;
 }
 
-void set_datatype_destroy(const struct datatype *dtype)
-{
-	datatype_free(dtype);
-}
-
 static struct error_record *time_unit_parse(const struct location *loc,
 					    const char *str, uint64_t *unit)
 {
